/*=============================================================================
Fullerene
  Copyright © Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003 - 2006
  -- License restrictions apply (GPL v2)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published
  by the Free Software Foundation; however ONLY version 2 of the License.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public
  License along with this program kit; if not, write to:
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
  Boston, MA  02110-1301  USA

  -- For details, see the file named "LICENSE.GPL2"
=============================================================================*/
/*
        Displays
*/
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#define RAD2DEG(a) ((double)(a) * 180 / M_PI)
#define DEG2RAD(a) ((double)(a) * M_PI / 180)
#define SLASH(v) (floor(v * 1000) / 1000)

//-----------------------------------------------------------------------------
#define GM 1.6180339887498948482045868343656381177203 // golden mean
static const double vertices[] = {
    /*  0 */           0,          -1,      3 * GM,
    /*  1 */          GM,          -2,  1 + 2 * GM,
    /*  2 */           1,     -2 - GM,      2 * GM,
    /*  3 */          -1,     -2 - GM,      2 * GM,
    /*  4 */         -GM,          -2,  1 + 2 * GM,
    /*  5 */           2, -1 - 2 * GM,          GM,
    /*  6 */           1,     -3 * GM,           0,
    /*  7 */          -1,     -3 * GM,           0,
    /*  8 */          -2, -1 - 2 * GM,          GM,
    /*  9 */           2, -1 - 2 * GM,         -GM,
    /* 10 */           1,     -2 - GM,     -2 * GM,
    /* 11 */          -1,     -2 - GM,     -2 * GM,
    /* 12 */          -2, -1 - 2 * GM,         -GM,
    /* 13 */          GM,          -2, -1 - 2 * GM,
    /* 14 */           0,          -1,     -3 * GM,
    /* 15 */         -GM,          -2, -1 - 2 * GM,
    /* 16 */      2 + GM,     -2 - GM,           1,
    /* 17 */  1 + 2 * GM,         -GM,           2,
    /* 18 */      2 * GM,          -1,      2 + GM,
    /* 19 */      2 + GM,     -2 - GM,          -1,
    /* 20 */     -2 - GM,     -2 - GM,          -1,
    /* 21 */     -2 - GM,     -2 - GM,           1,
    /* 22 */ -1 - 2 * GM,         -GM,           2,
    /* 23 */     -2 * GM,          -1,      2 + GM,
    /* 24 */  1 + 2 * GM,         -GM,          -2,
    /* 25 */      2 * GM,          -1,     -2 - GM,
    /* 26 */     -2 * GM,          -1,     -2 - GM,
    /* 27 */ -1 - 2 * GM,         -GM,          -2,
    /* 28 */     -3 * GM,           0,          -1,
    /* 29 */     -3 * GM,           0,           1,
    /* 30 */      3 * GM,           0,           1,
    /* 31 */      3 * GM,           0,          -1,
    /* 32 */      2 * GM,           1,      2 + GM,
    /* 33 */  1 + 2 * GM,          GM,           2,
    /* 34 */ -1 - 2 * GM,          GM,           2,
    /* 35 */     -2 * GM,           1,      2 + GM,
    /* 36 */         -GM,           2,  1 + 2 * GM, 
    /* 37 */           0,           1,      3 * GM, 
    /* 38 */          GM,           2,  1 + 2 * GM, 
    /* 39 */  1 + 2 * GM,          GM,          -2,
    /* 40 */      2 * GM,           1,     -2 - GM,
    /* 41 */      2 + GM,      2 + GM,           1,
    /* 42 */      2 + GM,      2 + GM,          -1,
    /* 43 */          GM,           2, -1 - 2 * GM,
    /* 44 */           0,           1,     -3 * GM, 
    /* 45 */         -GM,           2, -1 - 2 * GM,
    /* 46 */     -2 * GM,           1,     -2 - GM,
    /* 47 */ -1 - 2 * GM,          GM,          -2,
    /* 48 */     -2 - GM,      2 + GM,          -1,
    /* 49 */     -2 - GM,      2 + GM,           1,
    /* 50 */           2,  1 + 2 * GM,         -GM, 
    /* 51 */           1,      2 + GM,     -2 * GM, 
    /* 52 */           1,      2 + GM,      2 * GM,
    /* 53 */           2,  1 + 2 * GM,          GM, 
    /* 54 */          -1,      2 + GM,      2 * GM,
    /* 55 */          -2,  1 + 2 * GM,          GM, 
    /* 56 */          -1,      2 + GM,     -2 * GM, 
    /* 57 */          -2,  1 + 2 * GM,         -GM, 
    /* 58 */          -1,      3 * GM,           0,
    /* 59 */           1,      3 * GM,           0,
};
static const unsigned int index_penta[] = {
    /*  0 */  0,  1,  2,  3,  4,
    /*  5 */ 10, 13, 14, 15, 11,
    /* 10 */  5,  6,  9, 19, 16,
    /* 15 */  8,  7, 12, 20, 21,
    /* 20 */ 30, 17, 18, 32, 33,
    /* 25 */ 23, 22, 29, 34, 35,
    /* 30 */ 25, 24, 31, 39, 40,
    /* 35 */ 28, 27, 26, 46, 47,
    /* 40 */ 52, 38, 37, 36, 54,
    /* 45 */ 45, 44, 43, 51, 56,
    /* 50 */ 55, 49, 48, 57, 58,
    /* 55 */ 50, 42, 41, 53, 59,
};
static const unsigned int index_hexa[] = {
    /*   0 */  2,  5,  6,  7,  8,  3,
    /*   6 */  6,  9, 10, 11, 12,  7,
    /*  12 */  1,  2,  5, 16, 17, 18,
    /*  18 */  4,  3,  8, 21, 22, 23,
    /*  24 */ 13, 10,  9, 19, 24, 25,
    /*  30 */ 20, 12, 11, 15, 26, 27,
    /*  36 */ 22, 21, 20, 27, 28, 29,
    /*  42 */ 24, 19, 16, 17, 30, 31,
    /*  48 */  0,  4, 23, 35, 36, 37,
    /*  54 */ 32, 18,  1,  0, 37, 38,
    /*  60 */ 39, 31, 30, 33, 41, 42,
    /*  66 */ 14, 13, 25, 40, 43, 44,
    /*  72 */ 26, 15, 14, 44, 45, 46,
    /*  80 */ 34, 29, 28, 47, 48, 49,
    /*  88 */ 43, 40, 39, 42, 50, 51,
    /*  94 */ 41, 33, 32, 38, 52, 53,
    /* 100 */ 54, 36, 35, 34, 49, 55,
    /* 106 */ 48, 47, 46, 45, 56, 57,
    /* 112 */ 53, 52, 54, 55, 58, 59,
    /* 118 */ 57, 56, 51, 50, 59, 58,
};

// Functions
static void init_gl(int, int);
static void resize_glscene(int, int);
static void key_pressed(unsigned char, int, int);
static void draw_glscene(void);

static void d_quader(double, double, double, unsigned int);
static void d_rect(double, double, double, unsigned int);
static void d_shape(const unsigned int *, const double *, size_t, unsigned int);
static void rotate(double, double, double, double *, double *, double *, double *);
static void text_helv(const char *, double, double, double);

// Variables
static int window;
static int paused=0;
static double RH, RA, RB;
static int DrawBrickWF = 1, DrawBrick = 1, DrawShape = 0, DrawShapeWF = 1,
    DrawText = 1;

//-----------------------------------------------------------------------------
int main(int argc, const char **argv) {
    glutInit(&argc, (char **)argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA | GLUT_DEPTH);
    glutInitWindowSize(1024, 768);
    glutInitWindowPosition(0, 0);
    window = glutCreateWindow("Truncated Icosahedron (Fullerene 60)");
    glutDisplayFunc(&draw_glscene);
    glutIdleFunc(&draw_glscene);
    glutReshapeFunc(&resize_glscene);
    glutKeyboardFunc(&key_pressed);
    init_gl(640, 480);
    glutMainLoop();
    return 1;
}

static void init_gl(int width, int height) {
    glClearColor(255, 255, 255, 0);
    glClearDepth(1);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glShadeModel(GL_SMOOTH);
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable(GL_LINE_SMOOTH);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, (double)width / height, 0.1, 100);
    glMatrixMode(GL_MODELVIEW);

    srand(time(NULL) ^ getpid());
    return;
}

static void resize_glscene(int width, int height) {
    if(height == 0)
        height = 1;
    glViewport(0, 0, width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45, (double)width / height, 0.1, 100);
    glMatrixMode(GL_MODELVIEW);
    return;
}

static void key_pressed(unsigned char key, int x, int y) {
    static const int mod = 3;
    if(key == '\e') {
        glutDestroyWindow(window);
        exit(0);
    }
    else if(key == '4') RH -= mod;
    else if(key == '6') RH += mod;
    else if(key == '2') RA -= mod;
    else if(key == '8') RA += mod;
    else if(key == '9') RB += mod;
    else if(key == '3') RB -= mod;
    else if(key == ' ') paused = !paused;

    return;
}

static void draw_glscene(void) {
    double ra, rx, ry, rz;
    int cnt;

    if(paused) {
        usleep(100000);
        return;
    }

    glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);
    glLoadIdentity();
    glTranslated(0, 0, -15);
    glShadeModel(GL_SMOOTH);
    glEnable(GL_BLEND);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);

    rotate(DEG2RAD(RH), DEG2RAD(RA), DEG2RAD(RB), &ra, &rx, &ry, &rz);
    glRotated(ra, rx, ry, rz);

    // Golden mean
    glLineWidth(1.0 / GM);

    // Orthogonal rectangles
    if(DrawBrick) {
        glColor4f(0.5, 0.5, 1, .5);
        d_rect(0, 1, 3 * GM, GL_POLYGON);
        glColor4f(0.4, 0.4, 1, .5);
        d_rect(1, 3 * GM, 0, GL_POLYGON);
        glColor4f(0.6, 0.6, 1, .5);
        d_rect(3 * GM, 0, 1, GL_POLYGON);
    }
    if(DrawBrickWF) {
        glColor4f(0, 0, 1, .5);
        d_rect(0, 1, 3 * GM, GL_LINE_LOOP);
        d_rect(1, 3 * GM, 0, GL_LINE_LOOP);
        d_rect(3 * GM, 0, 1, GL_LINE_LOOP);
    }

    // Orthogonal bricks
    if(DrawBrick) {
        glColor4f(1, 0.5, 0.5, .5);
        d_quader(2, 1 + 2 * GM, GM, GL_POLYGON);
        glColor4f(1, 0.4, 0.4, .5);
        d_quader(1 + 2 * GM, GM, 2, GL_POLYGON);
        glColor4f(1, 0.6, 0.6, .5);
        d_quader(GM, 2, 1 + 2 * GM, GL_POLYGON);
    }
    if(DrawBrickWF) {
        glColor4f(1, 0, 0, .5);
        d_quader(2, 1 + 2 * GM, GM, GL_LINE_LOOP);
        d_quader(1 + 2 * GM, GM, 2, GL_LINE_LOOP);
        d_quader(GM, 2, 1 + 2 * GM, GL_LINE_LOOP);
    }

    if(DrawBrick) {
        glColor4f(0.5, 1, 0.5, .5);
        d_quader(1, 2 + GM, 2 * GM, GL_POLYGON);
        glColor4f(0.4, 1, 0.4, .5);
        d_quader(2 + GM, 2 + GM, 1, GL_POLYGON);
        glColor4f(0.6, 1, 0.6, .5);
        d_quader(2 * GM, 1, 2 + GM, GL_POLYGON);
    }
    if(DrawBrickWF) {
        glColor4f(0, 0.5, 0, .5);
        d_quader(1, 2 + GM, 2 * GM, GL_LINE_LOOP);
        d_quader(2 + GM, 2 + GM, 1, GL_LINE_LOOP);
        d_quader(2 * GM, 1, 2 + GM, GL_LINE_LOOP);
    }

    // Shapes
    glColor3ub(0,0,0);
    for(cnt = 0; cnt <= 55; cnt += 5) {
        if(DrawShape) {
            glColor4ub(0xff, 0xff, 0xff, 0x80);
            d_shape(&index_penta[cnt], vertices, 5, GL_POLYGON);
        }
        if(DrawShapeWF) {
            glColor4ub(0, 0, 0, 0xff);
            d_shape(&index_penta[cnt], vertices, 5, GL_LINE_LOOP);
        }
    }
    for(cnt = 0; cnt <= 118; cnt += 6) {
        if(DrawShape) {
            glColor4ub(0xff, 0xff, 0xff, 0x80);
            d_shape(&index_hexa[cnt], vertices, 6, GL_POLYGON);
        }
        if(DrawShapeWF) {
            glColor4ub(0, 0, 0, 0xff);
            d_shape(&index_hexa[cnt], vertices, 6, GL_LINE_LOOP);
        }
    }

    // swap the buffers to display, since double buffering is used.
    glutSwapBuffers();
    return;
}

//-----------------------------------------------------------------------------
static void d_quader(double x, double y, double z, unsigned int meth) {
    glBegin(meth); { // back
        glVertex3d(-x, -y, -z);
        glVertex3d( x, -y, -z);
        glVertex3d( x,  y, -z);
        glVertex3d(-x,  y, -z);
    }; glEnd();
    glBegin(meth); { // front
        glVertex3d(-x, -y, z);
        glVertex3d( x, -y, z);
        glVertex3d( x,  y, z);
        glVertex3d(-x,  y, z);
    }; glEnd();
    glBegin(meth); { // left
        glVertex3d(-x, -y,  z);
        glVertex3d(-x, -y, -z);
        glVertex3d(-x,  y, -z);
        glVertex3d(-x,  y,  z);
    }; glEnd();
    glBegin(meth); { // right
        glVertex3d(x, -y,  z);
        glVertex3d(x, -y, -z);
        glVertex3d(x,  y, -z);
        glVertex3d(x,  y,  z);
    }; glEnd();
    glBegin(meth); { // top
        glVertex3d(-x, -y, -z);
        glVertex3d( x, -y, -z);
        glVertex3d( x, -y,  z);
        glVertex3d(-x, -y,  z);
    }; glEnd();
    glBegin(meth); { // bottom
        glVertex3d(-x, y, -z);
        glVertex3d( x, y, -z);
        glVertex3d( x, y,  z);
        glVertex3d(-x, y,  z);
    }; glEnd();
    return;
}

static void d_rect(double x, double y, double z, unsigned int meth) {
    glBegin(meth);
    if(x == 0) {
        glVertex3d(0, -y, -z);
        glVertex3d(0,  y, -z);
        glVertex3d(0,  y,  z);
        glVertex3d(0, -y,  z);
    } else if(y == 0) {
        glVertex3d(-x, 0, -z);
        glVertex3d( x, 0, -z);
        glVertex3d( x, 0,  z);
        glVertex3d(-x, 0,  z);
    } else if(z == 0) {
        glVertex3d(-x, -y, 0);
        glVertex3d( x, -y, 0);
        glVertex3d( x,  y, 0);
        glVertex3d(-x,  y, 0);
    }
    glEnd();
    return;
}


static void d_shape(const unsigned int *idx, const double *pts, size_t n,
 unsigned int meth)
{
    const unsigned int *oidx = idx;
    size_t on = n;
    char buf[16];

    glBegin(meth);
    while(n--) {
        unsigned int b = *idx * 3;
        double x = pts[b], y = pts[b + 1], z = pts[b + 2];
        glVertex3d(x, y, z);
        ++idx;
    }
    glEnd();

    n = on;
    idx = oidx;
    while(n--) {
        unsigned int b = *idx * 3;
        double x = pts[b], y = pts[b + 1], z = pts[b + 2];
        snprintf(buf, sizeof(buf), "%d", *idx);
        if(DrawText)
            text_helv(buf, x, y, z);
        ++idx;
    }

    return;
}

static void rotate(double heading, double attitude, double bank, double *ap,
  double *xp, double *yp, double *zp)
{
    // Assuming the angles are in radians.
    double a, x, y, z, n,
     c1 = cos(heading/2), c2 = cos(attitude/2), c3 = cos(bank/2),
     s1 = sin(heading/2), s2 = sin(attitude/2), s3 = sin(bank/2),
     C1C2 = c1 * c2, S1S2 = s1 * s2,
     C1S2 = c1 * s2, S1C2 = s1 * c2;

    a = 2 * acos(C1C2 * c3 - S1S2 * s3);
    x = C1C2 * s3 + S1S2 * c3;
    y = S1C2 * c3 + C1S2 * s3;
    z = C1S2 * c3 - S1C2 * s3;

    if((n = x * x + y * y + z * z) < 0.00001) {
        x = z = 0;
        y = 1;
    } else {
        n = sqrt(n);
        x /= n;
        y /= n;
        z /= n;
    }
    *ap = RAD2DEG(SLASH(a));
    *xp = SLASH(x);
    *yp = SLASH(y);
    *zp = SLASH(z);
    return;
}

static void text_helv(const char *s, double x, double y, double z) {
    size_t len, i;
    glPushMatrix(); {
        glRasterPos3f(x, y, z);
        len = strlen(s);
        for(i = 0; i < len; ++i)
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, s[i]);
    } glPopMatrix();
    return;
}

//=============================================================================
